#ifndef DIRICHLETBC_RECTANGLE_H
#define DIRICHLETBC_RECTANGLE_H

#include "dirichletbc2d.h" 

/** DirichletBC_Rectangle:
 *      Assumes that the boundaries of the domain are the 4 edges of a rectangle.
 *      Each edge is either a dirichlet boundary or not (i.e. Dirichlet data cannot
 *      be given only on parts of an edge).
 */
class DirichletBC_Rectangle : public DirichletBC2D{

private:
    double xmin, xmax;      // The smallest/largest x-value of the domain
    double ymin, ymax;      // The smallest/largest y-value of the domain
    
    std::set<int> d_bc_ids; // The set of edges that are Dirichlet
                            // Boundary Ids:
                            //  0: bottom
                            //  1: right
                            //  2: top
                            //  3: left
public:
    /** Constructors
     */
    DirichletBC_Rectangle(){} 
    DirichletBC_Rectangle(Point2D& lower_left_corner, Point2D& upper_right_corner, std::set<int> _d_bc_ids);
    
    /** Returns whether x is on the Dirichlet boundary or not.
     */
    bool
    is_Dirichlet_point(Point2D& x); 


};

DirichletBC_Rectangle::DirichletBC_Rectangle(Point2D& lower_left_corner, 
                             Point2D& upper_right_corner, 
                             std::set<int> _d_bc_ids) :
    xmin(lower_left_corner.x), xmax(upper_right_corner.x),
    ymin(lower_left_corner.y), ymax(upper_right_corner.y),
    d_bc_ids(_d_bc_ids){}
   
bool
DirichletBC_Rectangle::is_Dirichlet_point(Point2D& x)
{

    // If x_2 == y_min and bottom edge is Dirichlet edge -> x on Dirichlet boundary
    if( fabs(x.y - ymin) < x.eps &&  d_bc_ids.find(0) != d_bc_ids.end()){
        return true;
    }
    // If x_1 == x_max and right edge is Dirichlet edge -> x on Dirichlet boundary
    if( fabs(x.x - xmax) < x.eps &&  d_bc_ids.find(1) != d_bc_ids.end()){
        return true;
    }
    // If x_2 == y_max and top edge is Dirichlet edge -> x on Dirichlet boundary
    if( fabs(x.y - ymax) < x.eps &&  d_bc_ids.find(2) != d_bc_ids.end()){        
        return true;
    }    
    // If x_1 == x_min and left edge is Dirichlet edge -> x on Dirichlet boundary
    if( fabs(x.x - xmin) < x.eps &&  d_bc_ids.find(3) != d_bc_ids.end()){
        return true;
    }

    // Else: x is not on Dirichlet boundary
    return false;
}

#endif // DIRICHLETBC_RECTANGLE_H