#ifndef ELLIPTICFEM2D_H
#define ELLIPTICFEM2D_H

#include <vector>
#include "sparsematrix.h"
#include "triangulation.h"
#include "pde2d.h"
#include "cg.h"
#include "dirichletbc2d.h" 

/** EllipticFEM2D:
 *      Provides structures for the assembly and solution 
 *      of a 2D stationary (symmetric) finite element problem
 *      on a given ordered mesh.
 */
class EllipticFEM2D{

private:
    Triangulation& mesh;			// The 2D triangular mesh
    std::vector<double> solution;	// The solution vector
    DirichletBC2D& d_bc;			// The object that specifies the Dirichlet boundary
    								//  (has a function is_dirichlet_point(x))
    
    SparseMatrix         A;			// System matrix A_h
    std::vector<double> F;			// Right hand side vector b_h
    
    PDE2D& pde2d;					// PDE Object specifying the problem to be solved
    
    /* Private Member Functions */

	/** Performs the assembly of A_h and b_h
     */
    void
    assemble();
    
public:
    /* Public Member Functions */
    
    /** Constructor
     */
    EllipticFEM2D(Triangulation& _mesh, PDE2D& _pde2d, DirichletBC2D& _d_bc);
    
	/** Assembles and solves the linear system
     */
    void
    solve();
    
    /** Prints both A_h and b_h to std::cout.
     *  Can be used to debug the system.
     */    
    void
    write_system();
    
    /** Write the solution to the specified file
     *  (in format: x_1 x_2 u(x))
     *  or print the vector to std::cout if no filename is given
     */
    void
    write_solution(std::string filename = std::string());
    
    /** Write the solution to the specified file
     *  (in grid format)
     *  or print the vector to std::cout if no filename is given
     */
    void
    write_solution_on_mesh(std::string filename = std::string());

};


//===========================================
// IMPLEMENTATIONS
//===========================================

/********
 * TODO * 
 ********/
EllipticFEM2D::EllipticFEM2D(Triangulation& _mesh, PDE2D& _pde2d, DirichletBC2D& _d_bc){}

/********
 * TODO * 
 ********/    
void
EllipticFEM2D::assemble()
{
    /* Performs the element-wise assembly by.
     * Handles the Dirichlet boundary conditions.
     */  
}

/********
 * TODO * 
 ********/
void
EllipticFEM2D::solve()
{
    /* Assembles and solves the system. 
     */
}

void
EllipticFEM2D::write_system()
{
    std::cout << "A = " << std::endl << A << std::endl;
    std::cout << "b = " << std::endl << F << std::endl;
}

void
EllipticFEM2D::write_solution(std::string filename)
{
    if(filename.size()>0){
        std::ofstream out(filename.c_str());
        if(out.good()){
            for(int i = 0; i < mesh.get_n_nodes(); ++i){
                out << mesh.get_node(i+1).x << " " << mesh.get_node(i+1).y << " " << solution[i] << std::endl;
            }
        }
        out.close();
        return;
    }
    std::cout << "u = " << std::endl << solution << std::endl;
}

void
EllipticFEM2D::write_solution_on_mesh(std::string filename)
{
    if(filename.size()>0){
        std::ofstream out(filename.c_str());
        if(out.good()){
            for (int i=1; i<= mesh.get_n_elements(); i++) {
                Polygon& p = mesh.get_element(i);
                for(int j = 0; j < p.N; ++j){
                     out << mesh.get_node(p.nodes[j]).x << " " << mesh.get_node(p.nodes[j]).y << " " << solution[p.nodes[j]-1] << std::endl;
                }
                out << mesh.get_node(p.nodes[0]).x << " " << mesh.get_node(p.nodes[0]).y << " " << solution[p.nodes[0]-1] << std::endl;
                out << std::endl<< std::endl;
            }
        }
        out.close();
        return;
    }
    
    for (int i=1; i<= mesh.get_n_elements(); i++) {
        Polygon& p = mesh.get_element(i);
        for(int j = 0; j < p.N; ++j){
             std::cout << mesh.get_node(p.nodes[j]).x << " " << mesh.get_node(p.nodes[j]).y << " " << solution[p.nodes[j]-1] << std::endl;
        }
        std::cout << mesh.get_node(p.nodes[0]).x << " " << mesh.get_node(p.nodes[0]).y << " " << solution[p.nodes[0]-1] << std::endl;
        std::cout << std::endl << std::endl;
    }
}

#endif // ELLIPTICFEM2D_H