#ifndef POISSON2D_H
#define POISSON2D_H

#include "densematrix.h"
#include "pde2d.h"

// Define type for 2D function (-> shorthand notation)
typedef double (*Fct2D)(Point2D&);

/** Poisson2D:
 *      Problem class that implements a 2D Laplace Operator
 *      
 *      Optionally with 
 *       - (non-homogeneous) Dirichlet boundary conditions
 *       - (non-homogeneous) initial values for non-stationary problems
 */
class Poisson2D : public PDE2D{
    
    /* Private Member Variables */
    
    Fct2D f;            // Right hand side function
    Fct2D d_bc;         // Dirichlet boundary condition function
    Fct2D u0_fct;       // Initial value condition function
    
    DenseMatrix K_1, K_2, K_3;       // The matrices storing the integrals over the derivatives
                                     // of the shape functions, constant on the reference element.
    
    /** Function to set the values of the K matrices
     */
    void
    initialize_K_matrizes();

public:
    
    /* Public Member Functions */
    
    /** Constructors
     */
    Poisson2D(Fct2D _f);
    Poisson2D(Fct2D _f, Fct2D _d_bc);
    Poisson2D(Fct2D _f, Fct2D _d_bc, Fct2D _u0_fct);

    /** Assembles A^(k) and b^(k) by transformation onto the reference element
     */
    void
    assemble_on_element(Point2D& x1, Point2D& x2, Point2D& x3, double Ak[3][3], double bk[3]);
    
    /** Returns the Dirichlet boundary condition at point x
     *  (overloads the PDE2D version)
     */
    double
    dirichlet_bc(Point2D& x);

    /** Returns the initial condition at point x.
     *  (overloads the PDE2D version)
     */
    double
    initial_condition(Point2D& x);

};

//===========================================
// IMPLEMENTATIONS
//===========================================


Poisson2D::Poisson2D(Fct2D _f) : 
    f(_f), K_1(3), K_2(3), K_3(3)
{
    initialize_K_matrizes();
}

Poisson2D::Poisson2D(Fct2D _f, Fct2D _d_bc) : 
    f(_f), d_bc(_d_bc), K_1(3), K_2(3), K_3(3)
{
    initialize_K_matrizes();
}

Poisson2D::Poisson2D(Fct2D _f, Fct2D _d_bc, Fct2D _u0_fct) : 
    f(_f), d_bc(_d_bc), u0_fct(_u0_fct), K_1(3), K_2(3), K_3(3)
{
    initialize_K_matrizes();
}

/********
 * TODO * 
 ********/
void
Poisson2D::initialize_K_matrizes()
{
    /* Fill K_1, K_2 and K_3 with the correct values
     */
}
 
/********
 * TODO * 
 ********/
void
Poisson2D::assemble_on_element(Point2D& x1, Point2D& x2, Point2D& x3, double Ak[3][3], double bk[3])
{
    /* Fill A and b with the correct values
     * using a transformation onto the reference element
     */
}

double
Poisson2D::dirichlet_bc(Point2D& x){
    // If d_bc exists, return function value
    if(d_bc != NULL){
        return d_bc(x);
    }
    // Else return default value 0
    return 0;
}

double
Poisson2D::initial_condition(Point2D& x){
    if(u0_fct != NULL){
        return u0_fct(x);
    }
    else{
        return 0;
    }
}

#endif // POISSON2D_H