#ifndef TRIANGULATION_H
#define TRIANGULATION_H

#include "point2d.h"
#include "polygon.h"
#include <vector>
#include <fstream>
#include <string>

#include <sstream>
#include <utility>
#include <map>
#include <algorithm>


template<class T> T fromString(const std::string& s)
{
     std::istringstream stream (s);
     T t;
     stream >> t;
     return t;
}


class Triangulation{
    std::vector<Point2D> coordinates; // stores the list of coordinates
    std::vector<Polygon> elements; // stores the list of elements

	
	//internal variables and functions for refinement!

	// This is a _unique_ list of edges (i.e. edge 1 2 is the same as 2 1) 
	// mapping to the storage location (=index) of the coordinates of its center point
	std::map<std::pair<int,int>,int> edge2newpoint;
	
	// This is a helper for the generation of the edge2newpoint map, 
	// it stores the number of the next new coordinate in the vector coordinates. I.e.
	// if a new edge is added to edge2newpoint, it maps to  coordinates_nextpos, 
	// then the center coordinates of the edge are added to the coordinate list and
	// coordinates_nextpos is increased by 1
	int coordinates_nextpos;
	
	// A list for the storage of the refined elements
	std::vector<Polygon> newelements;
	
	// A temporary variable that helps filling the list of new elements. 
	// As we do not want to use dynamic memory allocation of the vector class for each new element 
	// we initialize newelemens as 4*elements.size() and useelements_fillpos to get the first available position.
	int newelements_fillpos;

	//start refinement is used to initialize the internal variables 
	// (such as newelements etc)
	// that are used for refinement
	void 
	start_refinement();

	//finalize refinement combines the list elements 
	// with the vector newelements
	//do this with vectors of _fixed_ length, otherwise your program will run very long.
	void 
	finalize_refinement();

	//this refines the element at position element_nb into 4 Elements (using the variables above)
	void
	refine_red(int element_nb);

	//this returns a sorted pair of integers to make it easier to manage the edge2newpoint vector
	std::pair<int,int>
    sortedpair(const int a, const int b);


public:
	
    //Constructor: reads the files of the coordinates and the file of the elements
    Triangulation(const char* coordinatefile, const char* elementfile);

    void output(const char* outputfilename) const;

    void refineuniformly();

	//number of elements
    int
    get_n_elements();
    
	//number of nodes
    int
    get_n_nodes();
    
	//get a certain element
    Polygon&
    get_element(int nb);

    //get all nodes of an element
    std::vector<Point2D>
    get_element_nodes(int element_nb);
    
	//get a certain coordinate
    Point2D&
    get_node(int nb);
    
    
	void
    write_coordinates(const char* coordfile);
    
    void
    write_elements(const char* elemfile);

};


//Triangulation::Triangulation(){};

Triangulation::Triangulation(const char* coordinatefile, const char* elementfile){
    int linecount = 1;
    //int elemcount = 1;
    double a,b;
    //size_t pos;
    std::string line;
    std::ifstream coordfile(coordinatefile);
    if(coordfile.is_open()){
        while(coordfile.good()){
            getline(coordfile,line);
            std::string buffer;
            std::stringstream temp(line);
            std::vector<std::string> tokens;
            while(temp >> buffer){
                tokens.push_back(buffer);
            }
            if(tokens.size() != 2){
                std::cerr << "Error in file " << coordinatefile << " in line " << line << std::endl;
            }
            a = fromString<double>(tokens[0]);
            b = fromString<double>(tokens[1]);
            Point2D point(a,b);
            coordinates.push_back(point);
            linecount++;

        }
        coordfile.close();
    }

    linecount = 1;
    int i1,i2,i3;
    std::ifstream elemfile(elementfile);
    if(elemfile.is_open()){
        while(elemfile.good()){
            getline(elemfile,line);
           
            std::string buffer;
            std::stringstream temp(line);
            std::vector<std::string> tokens;
            while(temp >> buffer){
                tokens.push_back(buffer);
            }
            if(tokens.size() != 3){
                std::cerr << "Error in file " << elementfile << " in line " << line << std::endl;
            }
            i1 = fromString<int>(tokens[0]);
            i2 = fromString<int>(tokens[1]);
            i3 = fromString<int>(tokens[2]);
            
            
            Polygon poly(3);            
            poly[0] = i1;
            poly[1] = i2;
            poly[2] = i3;
            elements.push_back(poly);
            linecount++;
        }
        coordfile.close();
    }
    
}

std::pair<int,int>
Triangulation::sortedpair(const int a, const int b){
	return std::pair<int,int>(std::min(a,b),std::max(a,b));
}


/********
 * TODO * 
 ********/
void
Triangulation::start_refinement()
{}

/********
 * TODO * 
 ********/
void 
Triangulation::finalize_refinement()
{}

/********
 * TODO * 
 ********/
void Triangulation::refine_red(int element_nb)
{}

/********
 * TODO * 
 ********/
void
Triangulation::refineuniformly()
{}


void
Triangulation::output(const char* outputfilename) const{
    std::ofstream out(outputfilename);
    for (unsigned int i=0; i<elements.size(); i++) {
        for(int j = 0; j < elements[i].N; ++j){
            out << coordinates[elements[i].nodes[j]-1].x << " " << coordinates[elements[i].nodes[j]-1].y << std::endl;
        }
        out << coordinates[elements[i].nodes[0]-1].x << " " << coordinates[elements[i].nodes[0]-1].y << std::endl;
        out << std::endl;
    }
    out.close();
}

int
Triangulation::get_n_elements(){
    return elements.size();
}

int
Triangulation::get_n_nodes(){
    return coordinates.size();
}

Polygon&
Triangulation::get_element(int nb)
{
    return elements[nb-1];
}

std::vector<Point2D>
Triangulation::get_element_nodes(int element_nb)
{
    std::vector<Point2D> coords;
    
    coords.push_back(coordinates[elements[element_nb-1].nodes[0]-1]);
    coords.push_back(coordinates[elements[element_nb-1].nodes[1]-1]);
    coords.push_back(coordinates[elements[element_nb-1].nodes[2]-1]);
    
    return coords;
}

Point2D&
Triangulation::get_node(int nb)
{
    return coordinates[nb-1];
}


void
Triangulation::write_coordinates(const char* coordfile)
{
    std::ofstream cfile(coordfile);
    if(!cfile.is_open()){
        std::cerr << "Could not open file " << coordfile << " for writing!" << std::endl;
        exit(1);
    }
    for(std::size_t i = 0; i < coordinates.size(); ++i){
        cfile << coordinates[i].x << " " << coordinates[i].y << std::endl;
    }
    cfile.close();
}

void
Triangulation::write_elements(const char* elemfile)
{
    std::ofstream efile(elemfile);
    if(!efile.is_open()){
        std::cerr << "Could not open file " << elemfile << " for writing!" << std::endl;
        exit(1);
    }
    for(std::size_t i = 0; i < elements.size(); ++i){
        efile << elements[i][0] << " " << elements[i][1] << " " << elements[i][2] << std::endl;
    }
    efile.close();
}

#endif //TRIANGULATION_H